#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

#include "vncviewer.h"
#include "fbgl.h"
#include "overlay.h"

#if 0
#define DEBUG_CLIP
#endif

/***

New scaling code:

p_scale is integer from 1 to 32, equal to number of physical pixels per SCALE
virtual pixels. The real scaling is thus p_scale/SCALE.  Only the following scale
factors are currently supported:

	p_scale    physical/virtual
	1	=> 1/SCALE
	2	=> 2/SCALE
	3	=> 3/SCALE
	4	=> 4/SCALE
	5	=> 5/SCALE
	6	=> 6/SCALE
	7	=> 7/SCALE
	8	=> 8/SCALE
	16	=> 2
	24	=> 3
	32	=> 4

All operations are done on SCALExSCALE *virtual* pixel blocks, resulting in output
blocks of p_scale*p_scale pixels (between 1x1 and 32x32 pixels).

Invariants:

- v_x0 and v_y0 are always multiples of SCALE

Code redo:

old:
	virtual = physical * v_scale;
new:
	physical = virtual * p_scale / SCALE
	virtual = physical * SCALE / p_scale;

***/


void
vp_pan_virt(int x0, int y0) {
	int w, h;
	IMPORT_FRAMEBUFFER_VARS
	
	if (p_landscape) {
		w = p_xsize * SCALE / p_scale;
		h = p_ysize * SCALE / p_scale;
	} else {
		w = p_ysize * SCALE / p_scale;
		h = p_xsize * SCALE / p_scale;
	}

	if (v_xsize > w) {
		/* virtual screen wider than display */
		if (x0 + w > v_xsize) x0 = v_xsize - w;
		if (x0 < 0) x0 = 0;
	} else {
		/* width fits inside display */
		if (-x0 + v_xsize > w) x0 = v_xsize - w;
		if (x0 > 0) x0 = 0;
	}

	if (v_ysize > h) {
		/* virtual screen higher than display */
		if (y0 + h > v_ysize) y0 = v_ysize - h;
		if (y0 < 0) y0 = 0;
	} else {
		/* height fits inside display */
		if (-y0 + v_ysize > h) y0 = v_ysize - h;
		if (y0 > 0) y0 = 0;
	}

	/* ensure that x0 and y0 are multiples of SCALE */
	x0 &= ~(SCALE-1);
	y0 &= ~(SCALE-1);

	global_framebuffer.v_x0 = x0;
	global_framebuffer.v_y0 = y0;

	if ( 0 ) {
		/* HACK to tell window manager about new geometry */
		FILE *o;

		o=fopen("/tmp/fbwm.geom", "w");
		fprintf(o, "%d %d %d %d\n", x0, y0,
			v_xsize <= w ? v_xsize : w,
			v_ysize <= h ? v_ysize : h);
		fclose(o);

		system("killall -USR2 xfwm");
	}
	
	redraw_phys_all();
}

void
grid_pan(int dx, int dy) {
	int x0, y0;
	int gx, gy;
	IMPORT_FRAMEBUFFER_VARS

	if (p_landscape) {
		gx = p_xsize * SCALE / p_scale / 2;
		gy = p_ysize * SCALE / p_scale / 2;
	} else {
#if 0
		int tmp = dy;
		dy = -dx;
		dx = tmp;
#endif

		gx = p_ysize * SCALE / p_scale / 2;
		gy = p_xsize * SCALE / p_scale / 2;
	}

	if ((v_x0/gx)*gx == v_x0) {
		/* on grid */
		x0 = v_x0 + gx*dx;
	} else {
		/* snap to grid */
		x0 = (v_x0/gx)*gx;
		if (dx>0) x0 += gx;
	}
	if ((v_y0/gy)*gy == v_y0) {
		/* on grid */
		y0 = v_y0 + gy*dy;
	} else {
		/* snap to grid */
		y0 = (v_y0/gy)*gy;
		if (dy>0) y0 += gy;
	}

	if (p_landscape) {
		if (dx>0 && v_x0 + p_xsize * SCALE / p_scale >= v_xsize) x0=v_x0;
		if (dy>0 && v_y0 + p_ysize * SCALE / p_scale >= v_ysize) y0=v_y0;
	} else {
		if (dx>0 && v_x0 + p_ysize * SCALE / p_scale >= v_xsize) x0=v_x0;
		if (dy>0 && v_y0 + p_xsize * SCALE / p_scale >= v_ysize) y0=v_y0;
	}
	if (dx==0) x0=v_x0;
	if (dy==0) y0=v_y0;

	if (x0 != v_x0 || y0 != v_y0)
		vp_pan_virt(x0, y0);
}

void
vp_pan(int pdx, int pdy) {
	int x0, y0;
	IMPORT_FRAMEBUFFER_VARS

	if (p_landscape) {
		x0 = v_x0 + pdx*SCALE / p_scale;
		y0 = v_y0 + pdy*SCALE / p_scale;
	} else {
		x0 = v_x0 + pdy*SCALE / p_scale;
		y0 = v_y0 - pdx*SCALE / p_scale;
	}

	vp_pan_virt(x0, y0);
}

void flip_orientation() {
	List *p;
	IMPORT_FRAMEBUFFER_VARS
	
	for (p=global_framebuffer.overlays->next; p; p = p->next) {
		fbvnc_overlay_t *ov = p->val;

		if (p_landscape) {
			global_framebuffer.p_landscape = 0;
			ov->x = ov->x0p;
			ov->y = ov->y0p;
		} else {
			global_framebuffer.p_landscape = 1;
			ov->x = ov->x0l;
			ov->y = ov->y0l;
		}
	}
	
	vp_pan(0, 0);
}

/* pixel format: 
 *   1111110000000000
 *   5432109876543210
 *   rrrrrggggggbbbbb
 *
 * intermediate format:
 *   33222222222211111111110000000000
 *   10987654321098765432109876543210
 *   xxRRRRRRRRRRGGGGGGGGGGBBBBBBBBBB
 *
 * maximum value = (max_pixel_component) * (scale*scale)
 *               = 63 * 16 = 63 * 4 * 4
 *               = 1008
 *		 < 1024 (10 bits per component)
 */

#define RGBC(c) (((c) & 0x1f)<<1 | ((c)&0x7e0)<<5 | ((c)&0xf800)<<10)
#define CR(c) ((c)>>20)
#define CG(c) ((c)>>10&1023)
#define CB(c) ((c)&1023)

typedef void aa_line_func(Pixel*src, Pixel*dst, int wp);

void
s1aa_line(Pixel *src, Pixel *dst, int wp)
{
	memcpy(dst, src, wp * sizeof(Pixel));
}

/*******************************************
 *        xp=0                  xp=p_xsize *
 *       +-----------------------------+   *
 *   yp=0|     xp             xp+wp    |   *
 *       |   yp+----------------+      |   *
 *       |     | j=0  p0P0  v v |      |   *
 *       |     | j=1  p1P1  v v |      |   *
 *       |     +----------------+      |   *
 *       |            v0v1             |   *
 *   yp= |            V0V1             |   *
 *p_ysize|                             |   *
 *       +-----------------------------+   *
 *                                         *
 *******************************************/
void
s1aa_line_portrait(Pixel *src, Pixel *dst, int hp)
{
	int i;
	IMPORT_FRAMEBUFFER_VARS

	for(i=0; i<hp; i+=2) {
		CARD32 v0, v1, p0, p1;
		v0 = *(CARD32*)src;
		v1 = *(CARD32*)(src-v_xsize);

		p0 = ((v1&0xffff) << 16) | (v0 & 0xffff);
		p1 =  (v1 & 0xffff0000)  | (v0 >> 16);

		*(CARD32*)dst           = p0;
		*(CARD32*)(dst+p_xsize) = p1;

		src += 2;
		dst += 2*p_xsize;
	}
}

Pixel
s2aa_pixel(Pixel *src)
{
	CARD32 rgb, p0, p1;
	IMPORT_FRAMEBUFFER_VARS

	/* Dirty trick: read two pixels at once. 
	 * The order of halfwords within a word doesn't actually
	 * matter, because they are averaged anyway.
	 */
	p0 = *(CARD32*)src;
	p1 = *(CARD32*)(src+v_xsize);

	rgb = RGBC(p0&0xffff) + RGBC(p0>>16)
	    + RGBC(p1&0xffff) + RGBC(p1>>16);

	return (rgb>>3 & 0x1f) | (rgb>>7 & 0x7e0) | (rgb>>12 & 0xf800);
}

void
s2aa_line(Pixel *src, Pixel *dst, int wp)
{
	int i;
	IMPORT_FRAMEBUFFER_VARS
	
	for (i=0; i<wp; i++) {
		*dst = s2aa_pixel(src);
		src += 2;
		dst++;
	}
}

void
s2aa_line_portrait(Pixel *src, Pixel *dst, int hp)
{
	int i;
	IMPORT_FRAMEBUFFER_VARS

	for(i=0; i<hp; i++) {
		*dst = s2aa_pixel(src);
		src += 2;
		dst += p_xsize;
	}
}

Pixel
s4aa_pixel(Pixel *src)
{
	CARD32 rgb = 0;
	int js;
	IMPORT_FRAMEBUFFER_VARS

	for (js=0; js<4; js++) {
		CARD32 p0, p1;
		Pixel *buf = src + js*v_xsize;

		p0 = *(CARD32*)(buf);
		p1 = *(CARD32*)(buf+2);

		rgb += RGBC(p0&0xffff) + RGBC(p0>>16)
		     + RGBC(p1&0xffff) + RGBC(p1>>16);
	}

	return (rgb>>5 & 0x1f) | (rgb>>9 & 0x7e0) | (rgb>>14 & 0xf800);

}

void
s4aa_line(Pixel *src, Pixel *dst, int wp)
{
	int i;
	IMPORT_FRAMEBUFFER_VARS
	
	for (i=0; i<wp; i++) {
		*dst = s4aa_pixel(src);
		src += 4;
		dst++;
	}
}

void
s4aa_line_portrait(Pixel *src, Pixel *dst, int hp)
{
	int i;
	IMPORT_FRAMEBUFFER_VARS

	for(i=0; i<hp; i++) {
		*dst = s4aa_pixel(src);
		src += 4;
		dst += p_xsize;
	}
}

void
draw_border(int xp, int yp, int wp, int hp,
	int oxp, int oyp, int owp, int ohp)
{
	Pixel *buf;
	int j;
	IMPORT_FRAMEBUFFER_VARS

#ifdef DEBUG_CLIP
	fprintf(stderr, "border xp=%d yp=%d wp=%d hp=%d oxp=%d oyp=%d owp=%d ohp=%d\n",
		xp, yp, wp, hp,
		oxp, oyp, owp, ohp);
#endif

	if (!wp || !hp) return;

	if (xp < oxp) wp-=oxp-xp, xp=oxp;
	if (yp < oyp) hp-=oyp-yp, yp=oyp;

	if (xp+wp > oxp+owp) wp = oxp-xp+owp;
	if (yp+hp > oyp+ohp) hp = oyp-yp+ohp;

	buf = p_buf + yp*p_xsize + xp;

	for (j=0; j<hp; j++) {
		memset(buf, 0, wp*sizeof(Pixel));
		buf += p_xsize;
	}
}

/*******************************************
 *        xp=0                  xp=p_xsize *
 *       +-----------------------------+   *
 *   yp=0|     xp             xp+wp    |   *
 *       |   yp+----------------+      |   *
 *       |     | v   v   v   v  |      |   *
 *       |     |j=0 j=1  v   v  |      |   *
 *       |     | v   v   v   v  |      |   *
 *       |     | v   v   v   v  |      |   *
 *       |     | v   v   v   v  |      |   *
 *       |     +----------------+      |   *
 *   yp= |                             |   *
 *p_ysize|                             |   *
 *       +-----------------------------+   *
 *                                         *
 *******************************************/

/* block-based redraw */
void 
redraw_phys_landscape(int xp, int yp, int wp, int hp)
{
	int i, j;
	block_func_t *block_fn = 0;
	int xv, yv;
	Pixel *src, *dst;
	IMPORT_FRAMEBUFFER_VARS

	xv = v_x0 + xp*SCALE / p_scale;
	yv = v_y0 + yp*SCALE / p_scale;

	src = v_buf + yv*v_xsize + xv;
	dst = p_buf + yp*p_xsize + xp;

	switch(p_scale) {
	case 40: block_fn = scale_block_r0_40; break;
	case 30: block_fn = scale_block_r0_30; break;
	case 20: block_fn = scale_block_r0_20; break;
	case 10: block_fn = scale_block_r0_10; break;
	case 9: block_fn = scale_block_r0_9; break;
	case 8: block_fn = scale_block_r0_8; break;
	case 7: block_fn = scale_block_r0_7; break;
	case 6: block_fn = scale_block_r0_6; break;
	case 5: block_fn = scale_block_r0_5; break;
	case 4: block_fn = scale_block_r0_4; break;
	case 3: block_fn = scale_block_r0_3; break;
	case 2: block_fn = scale_block_r0_2; break;
	case 1: block_fn = scale_block_r0_1; break;

	default:
		cleanup_and_exit("bad scale", EXIT_ERROR);
	}

	for (j=0; j<hp/p_scale; j++) {
		for (i=0; i<wp/p_scale; i++) {
			block_fn(dst + i*p_scale, src + i*SCALE, p_xsize, v_xsize);
		}
		dst += p_xsize * p_scale;
		src += v_xsize * SCALE;
	}
}

void
redraw_phys_portrait(xp, yp, wp, hp) {
	int i, j;
	block_func_t *block_fn = 0;
	int xv, yv;
	Pixel *src, *dst;
	IMPORT_FRAMEBUFFER_VARS

	xv = v_x0 + yp*SCALE / p_scale;
	yv = v_y0 + (p_xsize-xp-wp-1)*SCALE / p_scale;

	src = v_buf + (yv + wp*SCALE / p_scale)*v_xsize + xv;
	dst = p_buf + yp*p_xsize + xp;

	switch(p_scale) {
	case 40: block_fn = scale_block_r1_40; break;
	case 30: block_fn = scale_block_r1_30; break;
	case 20: block_fn = scale_block_r1_20; break;
	case 10: block_fn = scale_block_r1_10; break;
	case 9: block_fn = scale_block_r1_9; break;
	case 8: block_fn = scale_block_r1_8; break;
	case 7: block_fn = scale_block_r1_7; break;
	case 6: block_fn = scale_block_r1_6; break;
	case 5: block_fn = scale_block_r1_5; break;
	case 4: block_fn = scale_block_r1_4; break;
	case 3: block_fn = scale_block_r1_3; break;
	case 2: block_fn = scale_block_r1_2; break;
	case 1: block_fn = scale_block_r1_1; break;

	default:
		cleanup_and_exit("bad scale", EXIT_ERROR);
	}

	for (j=0; j<wp/p_scale; j++) {
		for (i=0; i<hp/p_scale; i++) {
			block_fn(dst + i*p_xsize*p_scale, src + i*SCALE, p_xsize, v_xsize);
		}
		dst += p_scale;
		src -= SCALE*v_xsize;
	}
}

/* optimization opportunity:
 *
 * hextile redraws cause many 1x1 pixel screen updates that get inflated
 * to SCALExSCALE blocks, causing multiple redraws for adjacent pixels. It's still
 * fairly fast, but could be improved, for example with a width/SCALE x height/SCALE
 * bitmap to mark the areas needing redraw, then doing all just once.
 */

void
redraw_virt(int xv, int yv, int wv, int hv) {
	int xp, yp, hp, wp;
#ifdef DEBUG_CLIP
	int oxv, oyv, owv, ohv;
#endif
	IMPORT_FRAMEBUFFER_VARS

#ifdef DEBUG_CLIP
	fprintf(stderr, "redraw_virt: xv=%d yv=%d wv=%d hv=%d\n", xv, yv, wv, hv);
	oxv=xv; oyv=yv; owv=wv; ohv=hv;
#endif

	/* ensure coords and sizes are divisible by SCALE */
	/*printf("before xv:%d\n", xv);
	printf("before yv:%d\n", yv);
	printf("before wv:%d\n", wv);
	printf("before hv:%d\n", hv);*/
	{
		int dx, dy;
		dx = xv % SCALE;
		dy = yv % SCALE;

		xv -= dx;
		yv -= dy;
		wv += dx;
		hv += dy;
	}
	wv = ((wv+(SCALE-1)) / SCALE) * SCALE;
	hv = ((hv+(SCALE-1)) / SCALE) * SCALE;
	/*printf("after xv:%d\n", xv);
	printf("after yv:%d\n", yv);
	printf("after wv:%d\n", wv);
	printf("after hv:%d\n", hv);*/

#ifdef DEBUG_CLIP
	if (xv > oxv || xv+wv < oxv+owv || yv > oyv || yv+hv < oyv+ohv) 
		fprintf(stderr, "*** BAD! *** ");
	
	fprintf(stderr, "\trnd: xv=%d yv=%d wv=%d hv=%d\n", xv, yv, wv, hv);
#endif

	/* get corresponding physical coordinates */
	if (p_landscape) {
		xp = (xv - v_x0) * p_scale / SCALE;
		yp = (yv - v_y0) * p_scale / SCALE;
		wp = wv * p_scale / SCALE;
		hp = hv * p_scale / SCALE;
	} else {
		xp = p_xsize - (yv + hv - v_y0) * p_scale / SCALE;
		yp = (xv - v_x0) * p_scale / SCALE;
		wp = hv * p_scale / SCALE;
		hp = wv * p_scale / SCALE;
	}

#ifdef DEBUG_CLIP
	fprintf(stderr, "\tphys xp=%d yp=%d wp=%d hp=%d\n", xp, yp, wp, hp);
#endif

	/* clip to screen size */
	if (xp < 0) { wp += xp; xp = 0; }
	if (yp < 0) { hp += yp; yp = 0; }
	if (xp+wp > p_xsize) wp = p_xsize - xp;
	if (yp+hp > p_ysize) hp = p_ysize - yp;

	/* entirely off-screen ? */
	if (wp<=0 || hp<=0) return;

#ifdef DEBUG_CLIP
	fprintf(stderr, "\tclip xp=%d yp=%d wp=%d hp=%d\n", xp, yp, wp, hp);
#endif

	if (p_landscape) {
		redraw_phys_landscape(xp, yp, wp, hp);
	} else {
		redraw_phys_portrait(xp, yp, wp, hp);
	}
	if (!hide_overlays) redraw_overlays(xp, yp, wp, hp);
}

void
redraw_phys(int xp, int yp, int wp, int hp) {
	int xv, yv;
	int oxp=xp, oyp=yp, owp=wp, ohp=hp;
	IMPORT_FRAMEBUFFER_VARS

#ifdef DEBUG_CLIP
	fprintf(stderr, "redraw_phys: xp=%d yp=%d wp=%d hp=%d\n", xp, yp, wp, hp);
#endif

	if (p_landscape) {
		xv = v_x0 + xp*SCALE / p_scale;
		yv = v_y0 + yp*SCALE / p_scale;

		if (yv<0) {
			yp = -v_y0*p_scale/SCALE;
			yv = 0;
			draw_border(0, 0, p_xsize, yp,
				oxp, oyp, owp, ohp);
		}
		if (yv+hp*SCALE/p_scale > v_ysize) {
			hp = (v_ysize - yv)*p_scale/SCALE;
			draw_border(0, yp+hp, p_xsize, p_ysize-yp-hp,
				oxp, oyp, owp, ohp);
		}
		if (xv<0) {
			xp = -v_x0*p_scale/SCALE;
			xv = 0;
			draw_border(0, yp, xp, hp,
				oxp, oyp, owp, ohp);
		}
		if (xv+wp*SCALE/p_scale > v_xsize) {
			wp = (v_xsize - xv)*p_scale/SCALE;
			draw_border(xp+wp, yp, p_xsize-xp-wp, hp,
				oxp, oyp, owp, ohp);
		}
	} else {
		xv = v_x0 + yp*SCALE/p_scale;
		yv = v_y0 + (p_xsize-xp-wp)*SCALE/p_scale;

		if (xv<0) {
			yp = -v_x0*p_scale/SCALE;
			xv = 0;
			draw_border(0, 0, p_xsize, yp,
				oxp, oyp, owp, ohp);
		}
		if (xv+hp*SCALE/p_scale > v_xsize) {
			hp = (v_xsize - xv)*p_scale/SCALE;
			draw_border(0, yp+hp, p_xsize, p_ysize-yp-hp,
				oxp, oyp, owp, ohp);
		}
		if (yv<0) {
			int d = -yv * p_scale/SCALE;
			xp -= d;
			yv = 0;
			draw_border(xp+wp, yp, p_xsize-xp-wp, hp,
				oxp, oyp, owp, ohp);
		}
		if (yv+wp*SCALE/p_scale > v_ysize) {
			int d = wp - (v_ysize - yv)*p_scale/SCALE;
			wp = (v_ysize - yv)*p_scale/SCALE;
			xp += d;
			draw_border(0, yp, xp, hp,
				oxp, oyp, owp, ohp);
		}
	}

#ifdef DEBUG_CLIP
	fprintf(stderr, "\tclip xp=%d yp=%d wp=%d hp=%d\n", xp, yp, wp, hp);
#endif

	if (p_landscape) {
		redraw_phys_landscape(xp, yp, wp, hp);
	} else {
		redraw_phys_portrait(xp, yp, wp, hp);
	}
	if (!hide_overlays) redraw_overlays(oxp, oyp, owp, ohp);
}

void
redraw_phys_all(void)
{
	IMPORT_FRAMEBUFFER_VARS
	
	redraw_phys(0, 0, p_xsize, p_ysize);
	redraw_all_overlays();
}
