#ifndef _FBGL_H
#define _FBGL_H

#include "list.h"

typedef int bool;
typedef unsigned short Pixel;

struct viewport {
	int v_xsize;	/* total X size of virtual screen */
	int v_ysize;	/* total Y size of virtual screen */
	int v_x0;	/* x coord of phys screen's origin */
	int v_y0;	/* y coord of phys screen's origin */
	int v_bpp;	/* bits per pixel */
	Pixel *v_buf;

	bool p_landscape;
	int p_scale;	/* phys pixels per 8 virtual pixels,
	                 * see comments in fbdraw.c */
};

struct tsdev {
};

typedef struct {
	/* physical display */
	void *p_framebuf;
	int p_xsize;
	int p_ysize;
	bool p_landscape;
	int p_scale;	/* phys pixels per 8 virtual pixels,
	                 * see comments in fbdraw.c */
	Pixel *p_buf;

	/* viewport */
	int v_xsize;
	int v_ysize;
	int v_x0;
	int v_y0;
	int v_bpp;
	Pixel *v_buf;

	/* overlays */
	List *overlays;
	int hide_overlays; /* ref count */

	/* input devices */
	int ts_x;
	int ts_y;
	bool ts_pressed;

	struct tsdev *ts_dev;
	int kb_fd;

	/* user-defined input + output */
	int num_read_fds;
	int *read_fd;

	int num_write_fds;
	int *write_fd;
} fbvnc_framebuffer_t;

/* for efficiency reasons, don't access the contents of the global
 * structure directly. (Check the compiler's assembly
 * output - the compiler needs to reload the elements inside of loops,
 * because it cannot know if the elements might be modified by other
 * functions in the meantime.)
 */
extern fbvnc_framebuffer_t global_framebuffer;

/* The following macro defines auto variables with copies of the
 * global_framebufferstructure. The compiler is smart enough not to
 * load variables that aren't actually used. They are "const" to
 * guard against changing them accidentally - changes must be
 * made to the real global_framebuffer vars.
 */
#define IMPORT_FRAMEBUFFER_VARS						\
	void *const p_framebuf = global_framebuffer.p_framebuf;		\
	const int p_xsize = global_framebuffer.p_xsize;			\
	const int p_ysize = global_framebuffer.p_ysize;			\
	const bool p_landscape = global_framebuffer.p_landscape;	\
	Pixel *const p_buf = global_framebuffer.p_buf;			\
	const int v_xsize = global_framebuffer.v_xsize;			\
	const int v_ysize = global_framebuffer.v_ysize;			\
	const int v_x0 = global_framebuffer.v_x0;			\
	const int v_y0 = global_framebuffer.v_y0;			\
	const int p_scale = global_framebuffer.p_scale;			\
	const int v_bpp = global_framebuffer.v_bpp;			\
	Pixel *const v_buf = global_framebuffer.v_buf;			\
	List *const overlays = global_framebuffer.overlays;		\
	const bool hide_overlays = global_framebuffer.hide_overlays;	\
	const int ts_x = global_framebuffer.ts_x;			\
	const int ts_y = global_framebuffer.ts_y;			\
	const bool ts_pressed = global_framebuffer.ts_pressed;		\
	struct tsdev *const ts_dev = global_framebuffer.ts_dev;			\
	const int kb_fd = global_framebuffer.kb_fd;			\
	const int num_read_fds = global_framebuffer.num_read_fds;	\
	const int *read_fd = global_framebuffer.read_fd;		\
	const int num_write_fds = global_framebuffer.num_write_fds;	\
	int *const write_fd = global_framebuffer.write_fd;		\
/* Suppress 'unused variable' warnings */				\
	(void)p_framebuf;						\
	(void)p_xsize;							\
	(void)p_ysize;							\
	(void)p_landscape;						\
	(void)p_buf;							\
	(void)v_xsize;							\
	(void)v_ysize;							\
	(void)v_x0;							\
	(void)v_y0;							\
	(void)p_scale;							\
	(void)v_bpp;							\
	(void)v_buf;							\
	(void)overlays;							\
	(void)hide_overlays;						\
	(void)ts_x;							\
	(void)ts_y;							\
	(void)ts_pressed;						\
	(void)ts_dev;							\
	(void)kb_fd;							\
	(void)num_read_fds;						\
	(void)read_fd;							\
	(void)num_write_fds;						\
	(void)write_fd;							\
/* FRAMEBUFFER_VARS end */
	

enum fbvnc_event {
	FBVNC_EVENT_NULL = 0, /* false alarm */
	FBVNC_EVENT_TS_DOWN = 0x0001,
	FBVNC_EVENT_TS_MOVE = 0x0002,
	FBVNC_EVENT_TS_UP = 0x0004,

	FBVNC_EVENT_BTN_DOWN = 0x0010,
	FBVNC_EVENT_BTN_UP = 0x0020,

	FBVNC_EVENT_DATA_READABLE = 0x0100,
	FBVNC_EVENT_DATA_WRITABLE = 0x0200,

	FBVNC_EVENT_TIMEOUT = 0x1000,
	FBVNC_EVENT_SEND_UPDATE_REQUEST = 0x2000,
	FBVNC_EVENT_KEYREPEAT = 0x4000,
	FBVNC_EVENT_TICK_SECOND = 0x8000,
};

typedef struct {
	int evtype;
	int x, y;
	int dx, dy;
	int key;
	int fd;
} fbvnc_event_t;

extern List *sched;

typedef void block_func_t(Pixel *dst, Pixel *src, int xso, int xsi);

extern block_func_t scale_block_r0_1;
extern block_func_t scale_block_r0_2;
extern block_func_t scale_block_r0_3;
extern block_func_t scale_block_r0_4;
extern block_func_t scale_block_r0_5;
extern block_func_t scale_block_r0_6;
extern block_func_t scale_block_r0_7;
extern block_func_t scale_block_r0_8;
extern block_func_t scale_block_r0_9;
extern block_func_t scale_block_r0_10;
extern block_func_t scale_block_r0_20;
extern block_func_t scale_block_r0_30;
extern block_func_t scale_block_r0_40;

extern block_func_t scale_block_r1_1;
extern block_func_t scale_block_r1_2;
extern block_func_t scale_block_r1_3;
extern block_func_t scale_block_r1_4;
extern block_func_t scale_block_r1_5;
extern block_func_t scale_block_r1_6;
extern block_func_t scale_block_r1_7;
extern block_func_t scale_block_r1_8;
extern block_func_t scale_block_r1_9;
extern block_func_t scale_block_r1_10;
extern block_func_t scale_block_r1_20;
extern block_func_t scale_block_r1_30;
extern block_func_t scale_block_r1_40;


extern void schedule_add(List *s, int ms_delta, int event);

extern void gl_setpalettecolor(int i, int r, int g, int b);
extern void gl_copybox(int x1, int y1, int w, int h, int x2, int y2);
extern void gl_putbox(int x1, int y1, int w, int h, CARD8 *buf);
extern void gl_fillbox(int x, int y, int w, int h, int col);

#define EXIT_OK 0
#define EXIT_ERROR 1
#define EXIT_SYSERROR 2

extern void redraw_virt(int xv, int yv, int wv, int hv);
extern void redraw_phys(int xp, int yp, int wp, int hp);
extern void redraw_phys_all(void);
extern void vp_pan_virt(int x0, int y0);
extern void vp_pan(int pdx, int pdy);
extern void grid_pan(int dx, int dy);
extern void flip_orientation(void);
extern void cleanup_and_exit(char *msg, int ret);
extern void draw_border(int xp, int yp, int wp, int hp, int oxp, int oyp, int owp, int ohp);

extern int img_read(struct viewport *vport, FILE *f);

#define SCALE 10

#endif
