#!/usr/bin/perl

use strict;
$^W=1; # use warnings;

my $SB=10;

sub out (@) {
	print "\t", @_, ";\n";
}

sub outv($$) {
	my ($code, $fact) = @_;

	out $code, $fact if $fact;
}

sub out0($$$) {
	my ($code, $term, $fact) = @_;

	if ($fact) {
		out $code, $term, $fact;
	} else {
		out $code, 0;
	}
}

# C code generator
#
# Perl variables are uppercase, C vars lowercase

print "/* code auto-generated by $0, do not edit */\n\n";

for (my $S=1; $S <$SB; ++$S) {
	my $Xf = 0;
	my $Yf = 0;

	# need to use different conversion functions for uneven
	# scaling factors to avoid blue component overflow -
	# blue has only 10 bits, red and green have 11
	my $B = ($S&1) ? "b" : "";

	print "/* shrink 10:$S pixels, not rotated */
void
scale_block_r0_${S}(Pixel *dst, Pixel *src, int xso, int xsi)
{
	register Pix32 o, v;

";

	for (my $Y=0; $Y < $S; ++$Y) {
		out "register Pix32 t$Y = 0";
	}

	out "Pix32 u0 = 0" if $SB % $S;

	for (my $Y=0; $Y < $SB; ++$Y) {
		my $Ti = 0;

		$Yf += $S;

		for (my $X=0; $X < $SB; ++$X) {
			out "v = pix_16to32$B(src[$X])";

			$Xf += $S;

			if ($Xf < $SB) {
				if ($Yf < $SB) {
					outv "t$Ti += v*", $S*$S;
				} else {
					outv "t$Ti += v*", ($S-($Yf-$SB))*$S;
					outv "u0 += v*", ($Yf-$SB)*$S;
				}
			} else {
				$Xf -= $SB;

				if ($Yf < $SB) {
					outv "t$Ti += v*", ($S-$Xf)*$S;
					++ $Ti;
					outv "t$Ti += v*", $Xf*$S;
				} else {
					out "o=t$Ti + v*", ($S-($Yf-$SB))*($S-$Xf);
					out "dst[$Ti] = pix_32to16$B(o)";
					
					out0 "t$Ti = ", "u0 + v*", ($Yf-$SB)*($S-$Xf);
					++ $Ti;

					outv "t$Ti += v*", ($S-($Yf-$SB))*$Xf;
					out0 "u0 = ", "v*", ($Yf-$SB)*$Xf
						if $SB % $S;
				}
			}
			print "\n";
		}
		if ($Yf >= $SB) {
			$Yf -= $SB;
			out "dst += xso";
		}
		out "src += xsi";
		print "\n";
	}
	print "}\n\n";

}

for (my $S=$SB; $S <= $SB*4; $S += $SB) {
	print "/* enlarge 10:$S pixels, not rotated */
void
scale_block_r0_${S}(Pixel *dst, Pixel *src, int xso, int xsi)
{
	register Pixel p;

";

	my $F = $S/$SB;
	for (my $Y2=0; $Y2<$F; ++$Y2) {
		out "Pixel *d$Y2";
	}

	for (my $Y=0; $Y < $SB; ++$Y) {
		for (my $Y2=0; $Y2 < $F; $Y2++) {
			out "d$Y2 = dst";
			out "dst += xso";
		}
		print "\n";

		for (my $X=0; $X < $SB; ++$X) {
			out "p = src[$X]";
			for (my $Y2=0; $Y2 < $F; $Y2++) {
				for (my $X2=0; $X2 < $F; $X2++) {
					out "d$Y2", "[", $X*$F + $X2, "] = p";
				}
			}
		}
		out "src += xsi";
		print "\n";
	}
	print "}\n\n";
}

for (my $S=1; $S <$SB; ++$S) {
	my $Xf = 0;
	my $Yf = 0;

	my $B = ($S&1) ? "b" : "";

	print "/* shrink 10:$S pixels, rotated */
void
scale_block_r1_${S}(Pixel *dst, Pixel *src, int xso, int xsi)
{
	register Pix32 o, v;

";

	for (my $Y=0; $Y < $S; ++$Y) {
		out "Pix32 t$Y = 0";
	}

	out "register Pix32 u0 = 0" if $SB % $S;

	for (my $Y=0; $Y < $SB; ++$Y) {
		my $Ti = 0;

		$Yf += $S;

		for (my $X=0; $X < $SB; ++$X) {
			out "v = pix_16to32$B(src[-$X*xsi])";

			$Xf += $S;

			if ($Xf < $SB) {
				if ($Yf < $SB) {
					outv "t$Ti += v*", $S*$S;
				} else {
					outv "t$Ti += v*", ($S-($Yf-$SB))*$S;
					outv "u0 += v*", ($Yf-$SB)*$S;
				}
			} else {
				$Xf -= $SB;

				if ($Yf < $SB) {
					outv "t$Ti += v*", ($S-$Xf)*$S;
					++ $Ti;
					outv "t$Ti += v*", $Xf*$S;
				} else {
					out "o=t$Ti + v*", ($S-($Yf-$SB))*($S-$Xf);
					out "dst[$Ti] = pix_32to16$B(o)";
					
					out0 "t$Ti = ", "u0 + v*", ($Yf-$SB)*($S-$Xf);
					++ $Ti;

					outv "t$Ti += v*", ($S-($Yf-$SB))*$Xf;
					out0 "u0 = ", "v*", ($Yf-$SB)*$Xf
						if $SB % $S;
				}
			}
			print "\n";
		}
		if ($Yf >= $SB) {
			$Yf -= $SB;
			out "dst += xso";
		}
		out "src ++";
		print "\n";
	}
	print "}\n\n";

}

for (my $S=$SB; $S <= $SB*4; $S += $SB) {
	print "/* enlarge 10:$S pixels, rotated */
void
scale_block_r1_${S}(Pixel *dst, Pixel *src, int xso, int xsi)
{
	register Pixel p;

";

	my $F = $S/$SB;
	for (my $Y2=0; $Y2<$F; ++$Y2) {
		out "Pixel *d$Y2";
	}

	for (my $Y=0; $Y < $SB; ++$Y) {
		for (my $Y2=0; $Y2 < $F; $Y2++) {
			out "d$Y2 = dst";
			out "dst += xso";
		}
		print "\n";

		for (my $X=0; $X < $SB; ++$X) {
			out "p = src[-$X*xsi]";
			for (my $Y2=0; $Y2 < $F; $Y2++) {
				for (my $X2=0; $X2 < $F; $X2++) {
					out "d$Y2", "[", $X*$F + $X2, "] = p";
				}
			}
		}
		out "src ++";
		print "\n";
	}
	print "}\n\n";
}
